// Copyright (c) 2021 Terminus, Inc.
//
// This program is free software: you can use, redistribute, and/or modify
// it under the terms of the GNU Affero General Public License, version 3
// or later ("AGPL"), as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

package bundle

import (
	"strings"

	"github.com/pkg/errors"
	"github.com/sirupsen/logrus"

	"github.com/erda-project/erda/apistructs"
	"github.com/erda-project/erda/bundle/apierrors"
	"github.com/erda-project/erda/pkg/strutil"
)

func (b *Bundle) RunSoldierScript(scriptName string, params map[string]string) error {
	host, err := b.urls.Soldier()
	if err != nil {
		return err
	}
	hc := b.hc

	var runResp apistructs.RunScriptResponse
	req := hc.Post(host).Path(strutil.Concat("/autoop/run/", scriptName))
	for k, v := range params {
		req = req.Param(strutil.Concat("ENV_", k), v)
	}
	resp, err := req.Do().JSON(&runResp)
	if err != nil {
		return apierrors.ErrInvoke.InternalError(err)
	}
	if !resp.IsOK() || !runResp.Success {
		return toAPIError(resp.StatusCode(), runResp.Error)
	}

	if runResp.Data == false {
		return errors.Errorf("failed to run colony script, script: %s, soldier: %s", scriptName, host)
	}

	return nil
}

// MySQLInit 主从初始化
func (b *Bundle) MySQLInit(mysqlExecList *[]apistructs.MysqlExec, soldierUrl string) error {
	hc := b.hc
	logrus.Infof("MySQLInit body:%+v", *mysqlExecList)
	resp, err := hc.Post(soldierUrl).Path(apistructs.AddonMysqlInitURL).
		Header("Content-Type", "application/json").
		Header("Accept", "application/json").
		JSONBody(mysqlExecList).
		Do().DiscardBody()
	if err != nil {
		return apierrors.ErrInvoke.InternalError(err)
	}
	logrus.Infof("MySQLInit response :%+v", resp.StatusCode())
	if !resp.IsOK() {
		return errors.New("mysql初始化报错。")
	}
	return nil
}

// MySQLCheck 主从初始化
func (b *Bundle) MySQLCheck(mysqlExec *apistructs.MysqlExec, soldierUrl string) error {
	hc := b.hc
	logrus.Infof("MySQLCheck body:%+v", *mysqlExec)
	var checkResp apistructs.GetMysqlCheckResponse
	resp, err := hc.Post(soldierUrl).Path(apistructs.AddonMysqlcheckStatusURL).
		Header("Content-Type", "application/json").
		Header("Accept", "application/json").
		JSONBody(mysqlExec).
		Do().JSON(&checkResp)
	if err != nil {
		return apierrors.ErrInvoke.InternalError(err)
	}
	logrus.Infof("MySQLCheck response :%+v", *resp)
	if !resp.IsOK() {
		return errors.New("mysql主从状态同步检测报错。")
	}
	logrus.Infof("MySQLCheck checkResp :%+v", checkResp)
	if len(checkResp.Data) == 0 {
		return errors.New("mysql主从同步状态异常。")
	}
	for _, value := range checkResp.Data {
		if strings.ToLower(value) != "yes" && strings.ToLower(value) != "connecting" {
			return errors.New("mysql主从同步状态异常。")
		}
	}

	return nil
}

// MySQLExec 初始化mysql数据库
func (b *Bundle) MySQLExec(mysqlExec *apistructs.MysqlExec, soldierUrl string) error {
	hc := b.hc
	logrus.Infof("MySQLExec body: %+v", *mysqlExec)

	logrus.Infof("MySQLExec url: %s", soldierUrl)
	resp, err := hc.Post(soldierUrl).Path(apistructs.AddonMysqlExecURL).
		Header("Content-Type", "application/json").
		Header("Accept", "application/json").
		JSONBody(mysqlExec).
		Do().DiscardBody()
	if err != nil {
		return apierrors.ErrInvoke.InternalError(err)
	}
	logrus.Infof("MySQLExec response: %+v", *resp)
	if !resp.IsOK() {
		return errors.New("mysql初始化数据库报错。")
	}

	return nil
}

// MySQLExecFile 初始化mysql init.sql
func (b *Bundle) MySQLExecFile(mysqlExec *apistructs.MysqlExec, soldierUrl string) error {
	hc := b.hc
	logrus.Infof("executing init.sql, request body: %+v", *mysqlExec)
	logrus.Infof("executing init.sql url: %s", soldierUrl)
	resp, err := hc.Post(soldierUrl).Path(apistructs.AddonMysqlExecFileURL).
		Header("Content-Type", "application/json").
		Header("Accept", "application/json").
		JSONBody(mysqlExec).
		Do().DiscardBody()
	if err != nil {
		return apierrors.ErrInvoke.InternalError(err)
	}
	logrus.Infof("executed init.sql, response: %+v", *resp)
	if !resp.IsOK() {
		return errors.New("mysql init sql 报错。")
	}

	return nil
}
