// Copyright (c) 2021 Terminus, Inc.
//
// This program is free software: you can use, redistribute, and/or modify
// it under the terms of the GNU Affero General Public License, version 3
// or later ("AGPL"), as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

package testplan

import (
	"fmt"
	"reflect"
	"sort"
	"strconv"
	"time"

	"github.com/sirupsen/logrus"

	"github.com/erda-project/erda/apistructs"
	"github.com/erda-project/erda/bundle"
	"github.com/erda-project/erda/modules/qa/dao"
	"github.com/erda-project/erda/modules/qa/services/apierrors"
	"github.com/erda-project/erda/modules/qa/services/autotest"
	"github.com/erda-project/erda/modules/qa/services/testcase"
	"github.com/erda-project/erda/modules/qa/services/testset"
	"github.com/erda-project/erda/pkg/strutil"
)

// TestPlan
type TestPlan struct {
	db          *dao.DBClient
	bdl         *bundle.Bundle
	testCaseSvc *testcase.Service
	testSetSvc  *testset.Service
	autotest    *autotest.Service
}

// Option
type Option func(*TestPlan)

// New Initialize TestPlan service
func New(options ...Option) *TestPlan {
	svc := &TestPlan{}
	for _, op := range options {
		op(svc)
	}
	return svc
}

// WithDBClient Set db client
func WithDBClient(db *dao.DBClient) Option {
	return func(svc *TestPlan) {
		svc.db = db
	}
}

// WithBundle Set bundle
func WithBundle(bdl *bundle.Bundle) Option {
	return func(svc *TestPlan) {
		svc.bdl = bdl
	}
}

// WithTestCase Set testCaseSvc service
func WithTestCase(testCaseSvc *testcase.Service) Option {
	return func(svc *TestPlan) {
		svc.testCaseSvc = testCaseSvc
	}
}

// WithTestSet Set testSet service
func WithTestSet(testSetSvc *testset.Service) Option {
	return func(svc *TestPlan) {
		svc.testSetSvc = testSetSvc
	}
}

func WithAutoTest(autotest *autotest.Service) Option {
	return func(svc *TestPlan) {
		svc.autotest = autotest
	}
}

// Create 创建测试计划
func (t *TestPlan) Create(req apistructs.TestPlanCreateRequest) (uint64, error) {
	// req params check
	if req.Name == "" {
		return 0, apierrors.ErrCreateTestPlan.MissingParameter("name")
	}
	if req.OwnerID == "" {
		return 0, apierrors.ErrCreateTestPlan.MissingParameter("ownerID")
	}
	req.PartnerIDs = strutil.DedupSlice(req.PartnerIDs, true)
	if len(req.PartnerIDs) == 0 {
		return 0, apierrors.ErrCreateTestPlan.MissingParameter("partnerIDs")
	}
	if req.ProjectID == 0 {
		return 0, apierrors.ErrCreateTestPlan.MissingParameter("projectID")
	}

	if !req.IsInternalClient() {
		// Authorize
		access, err := t.bdl.CheckPermission(&apistructs.PermissionCheckRequest{
			UserID:   req.UserID,
			Scope:    apistructs.ProjectScope,
			ScopeID:  req.ProjectID,
			Resource: apistructs.TestPlanResource,
			Action:   apistructs.CreateAction,
		})
		if err != nil {
			return 0, err
		}
		if !access.Access {
			return 0, apierrors.ErrCreateTestPlan.AccessDenied()
		}
	}

	// create testplan
	testPlan := dao.TestPlan{
		Name:      req.Name,
		ProjectID: req.ProjectID,
		Status:    apistructs.TPStatusDoing,
		CreatorID: req.UserID,
		UpdaterID: req.UserID,
		Type:      apistructs.TestPlanTypeManual,
	}
	// 自动化测试计划
	if req.IsAutoTest {
		testPlan.Type = apistructs.TestPlanTypeAutoTest
		// 默认生成 inode
		inode, err := t.ensureAutoTestPlanInode(testPlan.ProjectID, testPlan.Name, req.IdentityInfo)
		if err != nil {
			return 0, apierrors.ErrCreateTestPlan.InternalError(err)
		}
		testPlan.Inode = inode
	}
	if err := t.db.CreateTestPlan(&testPlan); err != nil {
		return 0, apierrors.ErrCreateTestPlan.InternalError(err)
	}

	// create member
	var members []dao.TestPlanMember
	members = append(members, dao.TestPlanMember{
		TestPlanID: uint64(testPlan.ID),
		Role:       apistructs.TestPlanMemberRoleOwner,
		UserID:     req.OwnerID,
	})
	for _, partnerID := range req.PartnerIDs {
		members = append(members, dao.TestPlanMember{
			TestPlanID: uint64(testPlan.ID),
			Role:       apistructs.TestPlanMemberRolePartner,
			UserID:     partnerID,
		})
	}
	if err := t.db.BatchCreateTestPlanMembers(members); err != nil {
		return 0, apierrors.ErrCreateTestPlanMember.InternalError(err)
	}

	return uint64(testPlan.ID), nil
}

// Update 更新测试计划
func (t *TestPlan) Update(req apistructs.TestPlanUpdateRequest) error {
	// 参数校验
	req.PartnerIDs = strutil.DedupSlice(req.PartnerIDs, true)
	if req.Status != "" {
		if !req.Status.Valid() {
			return apierrors.ErrUpdateTestPlan.InvalidParameter(fmt.Sprintf("status: %s", req.Status))
		}
	}

	// 查询测试计划
	testPlan, err := t.db.GetTestPlan(req.TestPlanID)
	if err != nil {
		return apierrors.ErrUpdateTestPlan.InternalError(err)
	}
	if testPlan == nil {
		return apierrors.ErrUpdateTestPlan.NotFound()
	}

	// 更新测试计划
	if req.Name != "" {
		testPlan.Name = req.Name
	}
	if req.Status != "" {
		testPlan.Status = req.Status
	}
	if req.Summary != "" {
		testPlan.Summary = req.Summary
	}
	if req.TimestampSecStartedAt != nil {
		t := time.Unix(int64(*req.TimestampSecStartedAt), 0)
		testPlan.StartedAt = &t
	}
	if req.TimestampSecEndedAt != nil {
		t := time.Unix(int64(*req.TimestampSecEndedAt), 0)
		testPlan.EndedAt = &t
	}
	if err := t.db.UpdateTestPlan(testPlan); err != nil {
		return apierrors.ErrUpdateTestPlan.InternalError(err)
	}

	if req.OwnerID != "" || len(req.PartnerIDs) > 0 {
		members, err := t.db.ListTestPlanMembersByPlanID(req.TestPlanID)
		if err != nil {
			return apierrors.ErrListTestPlanMembers.InternalError(err)
		}
		var currentOwnerID string
		var currentPartnerIDs []string
		for _, mem := range members {
			if mem.Role.IsOwner() {
				currentOwnerID = mem.UserID
			}
			if mem.Role.IsPartner() {
				currentPartnerIDs = append(currentPartnerIDs, mem.UserID)
			}
		}

		// ownerID 不同时需要更新
		if req.OwnerID != "" && req.OwnerID != currentOwnerID {
			if err := t.db.OverwriteTestPlanOwner(req.TestPlanID, req.OwnerID); err != nil {
				return apierrors.ErrUpdateTestPlanMember.InternalError(err)
			}
		}

		// partnerIDs 不同时需要更新
		if len(req.PartnerIDs) > 0 {
			sort.Strings(currentPartnerIDs)
			sort.Strings(req.PartnerIDs)
			if !reflect.DeepEqual(currentPartnerIDs, req.PartnerIDs) {
				if err := t.db.OverwriteTestPlanPartners(req.TestPlanID, req.PartnerIDs); err != nil {
					return apierrors.ErrUpdateTestPlanMember.InternalError(err)
				}
			}
		}
	}

	return nil
}

// Delete
func (t *TestPlan) Delete(identityInfo apistructs.IdentityInfo, testPlanID uint64) error {
	testPlan, err := t.db.GetTestPlan(testPlanID)
	if err != nil {
		return err
	}
	if testPlan == nil {
		return apierrors.ErrDeleteTestPlan.NotFound()
	}

	if !identityInfo.IsInternalClient() {
		// Authorize
		access, err := t.bdl.CheckPermission(&apistructs.PermissionCheckRequest{
			UserID:   identityInfo.UserID,
			Scope:    apistructs.ProjectScope,
			ScopeID:  testPlan.ProjectID,
			Resource: apistructs.TestPlanResource,
			Action:   apistructs.DeleteAction,
		})
		if err != nil {
			return err
		}
		if !access.Access {
			return apierrors.ErrDeleteTestPlan.AccessDenied()
		}
	}

	// TODO 解除用例关联关系

	return t.db.DeleteTestPlan(testPlanID)
}

// Get 测试计划详情
func (t *TestPlan) Get(testPlanID uint64) (*apistructs.TestPlan, error) {
	testPlan, err := t.db.GetTestPlan(testPlanID)
	if err != nil {
		return nil, apierrors.ErrGetTestPlan.InternalError(err)
	}
	if testPlan == nil {
		return nil, apierrors.ErrGetTestPlan.NotFound()
	}
	// list member
	members, err := t.db.ListTestPlanMembersByPlanID(testPlanID)
	if err != nil {
		return nil, apierrors.ErrListTestPlanMembers.InternalError(err)
	}
	// list rels
	countMap, err := t.db.ListTestPlanCaseRelsCount([]uint64{testPlanID})
	if err != nil {
		return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(err)
	}

	convertedTp := t.Convert(testPlan, countMap[testPlanID], members...)

	return &convertedTp, nil
}

// List 测试计划分页查询
func (t *TestPlan) Paging(req apistructs.TestPlanPagingRequest) (*apistructs.TestPlanPagingResponseData, error) {
	// 参数校验
	if req.ProjectID == 0 {
		return nil, apierrors.ErrPagingTestPlans.MissingParameter("projectID")
	}
	if req.PageNo == 0 {
		req.PageNo = 1
	}
	if req.PageSize == 0 {
		req.PageSize = 20
	}
	for _, status := range req.Statuses {
		if !status.Valid() {
			return nil, apierrors.ErrPagingTestPlans.InvalidParameter(fmt.Errorf("status: %s", status))
		}
	}

	// paging testplan
	total, list, err := t.db.PagingTestPlan(req)
	if err != nil {
		return nil, apierrors.ErrPagingTestPlans.InternalError(err)
	}

	// list relation partners
	var tpIDs []uint64
	for _, tp := range list {
		tpIDs = append(tpIDs, uint64(tp.ID))
	}
	memberMap, err := t.db.ListTestPlanMembersByPlanIDs(tpIDs)
	if err != nil {
		return nil, apierrors.ErrListTestPlanMembers.InternalError(err)
	}

	// list rels count
	relCountMap, err := t.db.ListTestPlanCaseRelsCount(tpIDs)
	if err != nil {
		return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(err)
	}

	testPlans := make([]apistructs.TestPlan, 0, len(list))
	var userIDs []string
	for _, tp := range list {
		convertedTp := t.Convert(&tp, relCountMap[uint64(tp.ID)], memberMap[uint64(tp.ID)]...)
		testPlans = append(testPlans, convertedTp)
		userIDs = append(append(userIDs, convertedTp.OwnerID, convertedTp.CreatorID, convertedTp.UpdaterID), convertedTp.PartnerIDs...)
	}
	userIDs = strutil.DedupSlice(userIDs, true)

	return &apistructs.TestPlanPagingResponseData{
		Total:   total,
		List:    testPlans,
		UserIDs: userIDs,
	}, nil
}

// PagingTestPlanCaseRels 分页查询测试计划内测试用例
func (t *TestPlan) PagingTestPlanCaseRels(req apistructs.TestPlanCaseRelPagingRequest) (*apistructs.TestPlanCasePagingResponseData, error) {
	if req.TestPlanID == 0 {
		return nil, apierrors.ErrPagingTestPlanCaseRels.MissingParameter("testPlanID")
	}
	if req.PageNo == 0 {
		req.PageNo = 1
	}
	if req.PageSize == 0 {
		req.PageSize = 20
	}

	// 查询测试计划
	tp, err := t.Get(req.TestPlanID)
	if err != nil {
		return nil, err
	}

	// 从关系表根据过滤条件先过滤一次，用 relIDs 作为 测试用例分页查询全量列表
	// 更新时间起始时间 左闭区间过滤
	if req.TimestampSecUpdatedAtBegin != nil {
		t := time.Unix(int64(*req.TimestampSecUpdatedAtBegin), 0)
		req.UpdatedAtBeginInclude = &t
	}
	// 更新时间结束时间 右闭区间过滤
	if req.TimestampSecUpdatedAtEnd != nil {
		t := time.Unix(int64(*req.TimestampSecUpdatedAtEnd), 0)
		req.UpdatedAtEndInclude = &t
	}
	// 查询测试集包含子测试集
	allTestSetIDs, _, err := t.db.ListTestSetsRecursive(apistructs.TestSetListRequest{
		Recycled:      false,
		ParentID:      &req.TestSetID,
		ProjectID:     &tp.ProjectID,
		NoSubTestSets: false,
	})
	rels, err := t.db.ListTestPlanCaseRels(apistructs.TestPlanCaseRelListRequest{
		IDs:                   req.RelIDs,
		TestPlanIDs:           []uint64{req.TestPlanID},
		TestSetIDs:            allTestSetIDs,
		UpdaterIDs:            req.UpdaterIDs,
		ExecutorIDs:           req.ExecutorIDs,
		ExecStatuses:          req.ExecStatuses,
		UpdatedAtBeginInclude: req.UpdatedAtBeginInclude,
		UpdatedAtEndInclude:   req.UpdatedAtEndInclude,
		IdentityInfo:          req.IdentityInfo,
	})
	if err != nil {
		return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(err)
	}
	if len(rels) == 0 {
		return &apistructs.TestPlanCasePagingResponseData{
			Total:    0,
			TestSets: nil,
		}, nil
	}
	var relTcIDs []uint64
	testSetCaseMap := map[uint64][]uint64{}        // key: setID，value: [caseID]
	relMap := make(map[uint64]dao.TestPlanCaseRel) // key: tcID, value: rel
	for _, rel := range rels {
		relTcIDs = append(relTcIDs, rel.TestCaseID)
		relMap[rel.TestCaseID] = rel
		testSetCaseMap[rel.TestSetID] = append(testSetCaseMap[rel.TestSetID], rel.TestCaseID)
	}
	// 从测试用例获取用例相关数据
	pagingResult, err := t.testCaseSvc.PagingTestCases(apistructs.TestCasePagingRequest{
		ProjectID:      tp.ProjectID,
		TestSetID:      req.TestSetID,
		TestCaseIDs:    relTcIDs,
		TestSetCaseMap: testSetCaseMap,
		Query:          req.Query,
		Priorities:     req.Priorities,
		UpdaterIDs:     req.UpdaterIDs,
		//UpdatedAtBeginInclude: req.UpdatedAtBeginInclude,
		//UpdatedAtEndInclude:   req.UpdatedAtEndInclude,
		OrderByPriorityAsc:   req.OrderByPriorityAsc,
		OrderByPriorityDesc:  req.OrderByPriorityDesc,
		OrderByUpdaterIDAsc:  req.OrderByUpdaterIDAsc,
		OrderByUpdaterIDDesc: req.OrderByUpdaterIDDesc,
		OrderByUpdatedAtAsc:  req.OrderByUpdatedAtAsc,
		OrderByUpdatedAtDesc: req.OrderByUpdatedAtDesc,
		OrderByIDAsc:         req.OrderByIDAsc,
		OrderByIDDesc:        req.OrderByIDDesc,
		PageNo:               req.PageNo,
		PageSize:             req.PageSize,
	})
	if err != nil {
		return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(err)
	}

	// 组装新的数据结构，TestCase -> TestPlanCaseRel
	var newTestSets []apistructs.TestSetWithPlanCaseRels
	for _, ts := range pagingResult.TestSets {
		newTestSet := apistructs.TestSetWithPlanCaseRels{
			TestSetID: ts.TestSetID,
			Directory: ts.Directory,
			TestCases: nil,
		}
		for _, tc := range ts.TestCases {
			rel, ok := relMap[tc.ID]
			if !ok {
				return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(fmt.Errorf("not found testcase in testPlanCaseRelation, testCaseID: %d", tc.ID))
			}
			newRel := t.ConvertRel(&rel, &tc)
			newTestSet.TestCases = append(newTestSet.TestCases, *newRel)
			pagingResult.UserIDs = append(pagingResult.UserIDs, rel.CreatorID, rel.UpdaterID, rel.ExecutorID)
		}
		// inject total test case count without filter
		allRelsWithoutFilter, err := t.db.ListTestPlanCaseRels(apistructs.TestPlanCaseRelListRequest{
			TestPlanIDs: []uint64{tp.ID},
			TestSetIDs:  []uint64{ts.TestSetID},
			IDOnly:      true,
		})
		if err != nil {
			return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(fmt.Errorf("failed to get all rels count under testSet, testSetID: %d, err: %v", ts.TestSetID, err))
		}
		newTestSet.TestCaseCountWithoutFilter = uint64(len(allRelsWithoutFilter))
		newTestSets = append(newTestSets, newTestSet)
	}

	return &apistructs.TestPlanCasePagingResponseData{
		Total:    pagingResult.Total,
		TestSets: newTestSets,
		UserIDs:  strutil.DedupSlice(pagingResult.UserIDs),
	}, nil
}

func (t *TestPlan) ListTestPlanCaseRels(req apistructs.TestPlanCaseRelListRequest) (rels []apistructs.TestPlanCaseRel, err error) {
	defer func() {
		if r := recover(); r != nil {
			logrus.Errorf("failed to list test plan case rels, err: %v", r)
			err = apierrors.ErrPagingTestPlanCaseRels.InternalError(fmt.Errorf("invalid request"))
		}
	}()
	dbRels, err := t.db.ListTestPlanCaseRels(req)
	if err != nil {
		return nil, apierrors.ErrPagingTestPlanCaseRels.InternalError(err)
	}
	// get testcases
	tcMap := make(map[uint64]*apistructs.TestCase)
	if !req.IDOnly {
		var tcIDs []uint64
		for _, dbRel := range dbRels {
			tcIDs = append(tcIDs, dbRel.TestCaseID)
		}
		tcIDs = strutil.DedupUint64Slice(tcIDs, true)
		tcs, _, err := t.testCaseSvc.ListTestCases(apistructs.TestCaseListRequest{
			IDs:                   tcIDs,
			AllowMissingProjectID: true,
			AllowEmptyTestSetIDs:  true,
			Recycled:              false,
			IDOnly:                false,
		})
		if err != nil {
			return nil, err
		}
		if len(tcs) == 0 {
			return rels, nil
		}
		for i := range tcs {
			tcMap[tcs[i].ID] = &tcs[i]
		}
	}
	// assign tc to rel
	for _, dbRel := range dbRels {
		rel := t.ConvertRel(&dbRel, tcMap[dbRel.TestCaseID])
		rels = append(rels, *rel)
	}
	return rels, nil
}

// ExecuteAPITest 执行接口测试
func (t *TestPlan) ExecuteAPITest(req apistructs.TestPlanAPITestExecuteRequest) (uint64, error) {
	if req.TestPlanID == 0 {
		return 0, apierrors.ErrTestPlanExecuteAPITest.MissingParameter("testPlanID")
	}
	//if req.EnvID == 0 {
	//	return 0, apierrors.ErrTestPlanExecuteAPITest.MissingParameter("envID")
	//}

	// query test plan
	tp, err := t.Get(req.TestPlanID)
	if err != nil {
		return 0, err
	}

	// 获取测试用例列表
	if len(req.TestCaseIDs) == 0 {
		rels, err := t.db.ListTestPlanCaseRels(apistructs.TestPlanCaseRelListRequest{
			TestPlanIDs:  []uint64{tp.ID},
			IDOnly:       false,
			IdentityInfo: req.IdentityInfo,
		})
		if err != nil {
			return 0, apierrors.ErrTestPlanExecuteAPITest.InternalError(err)
		}
		for _, rel := range rels {
			req.TestCaseIDs = append(req.TestCaseIDs, rel.TestCaseID)
		}
	}

	// 调用 qa api 执行 api 测试
	qaAPITestReq := apistructs.ApiTestsActionRequest{
		ProjectID:        int64(tp.ProjectID),
		TestPlanID:       int64(req.TestPlanID),
		ProjectTestEnvID: int64(req.EnvID),
		UsecaseIDs:       req.TestCaseIDs,
	}

	return t.testCaseSvc.ExecuteAPIs(qaAPITestReq)
}

// Convert
func (t *TestPlan) Convert(testPlan *dao.TestPlan, relsCount apistructs.TestPlanRelsCount, members ...dao.TestPlanMember) apistructs.TestPlan {
	result := apistructs.TestPlan{
		ID:        uint64(testPlan.ID),
		Name:      testPlan.Name,
		Status:    testPlan.Status,
		ProjectID: testPlan.ProjectID,
		CreatorID: testPlan.CreatorID,
		UpdaterID: testPlan.UpdaterID,
		CreatedAt: &testPlan.CreatedAt,
		UpdatedAt: &testPlan.UpdatedAt,
		Summary:   testPlan.Summary,
		StartedAt: testPlan.StartedAt,
		EndedAt:   testPlan.EndedAt,
		RelsCount: relsCount,
		Type:      testPlan.Type,
		Inode:     testPlan.Inode,
	}
	for _, mem := range members {
		if mem.Role.IsOwner() {
			result.OwnerID = mem.UserID
		}
		if mem.Role.IsPartner() {
			result.PartnerIDs = append(result.PartnerIDs, mem.UserID)
		}
	}
	return result
}

// ensureAutoTestPlanInode 保证创建 autotest 计划的节点
// 测试计划也使用目录树进行保存，根目录下直接挂该项目下的叶子节点，即计划的 pipeline
func (t *TestPlan) ensureAutoTestPlanInode(projectID uint64, planName string, identityInfo apistructs.IdentityInfo) (string, error) {
	scopeID := strconv.FormatUint(projectID, 10)
	// 查询根目录，若根节点不存在，则自动创建
	rootNodes, err := t.autotest.ListFileTreeNodes(apistructs.UnifiedFileTreeNodeListRequest{
		Scope:        string(apistructs.AutoTestsScopeProjectTestPlan),
		ScopeID:      scopeID,
		Pinode:       "0",
		IdentityInfo: identityInfo,
	})
	if err != nil {
		return "", err
	}
	// 获取根节点
	var rootNodeInode string
	if len(rootNodes) > 0 {
		rootNodeInode = rootNodes[0].Inode
	} else {
		// 创建根目录
		newRootNode, err := t.autotest.CreateFileTreeNode(apistructs.UnifiedFileTreeNodeCreateRequest{
			Type:         apistructs.UnifiedFileTreeNodeTypeDir,
			Scope:        string(apistructs.AutoTestsScopeProjectTestPlan),
			ScopeID:      scopeID,
			Name:         "testplan-root-dir",
			IdentityInfo: identityInfo,
		})
		if err != nil {
			return "", err
		}
		rootNodeInode = newRootNode.Inode
	}
	// 创建子文件节点
	planNode, err := t.autotest.CreateFileTreeNode(apistructs.UnifiedFileTreeNodeCreateRequest{
		Type:         apistructs.UnifiedFileTreeNodeTypeFile,
		Scope:        string(apistructs.AutoTestsScopeProjectTestPlan),
		ScopeID:      scopeID,
		Pinode:       rootNodeInode,
		Name:         planName,
		IdentityInfo: identityInfo,
	})
	if err != nil {
		return "", err
	}
	return planNode.Inode, nil
}
